@@ -53,14 +53,12 @@ class InjectArgs(object):
 
     def __or__(self, other) -> Union[chain, Signature]:
         if isinstance(other, InjectArgs):
-            self.injectArgs.update(other.injectArgs)
+            r = InjectArgs(**(self.injectArgs | other.injectArgs))
         else:
+            r = other.clone()
             # chains and signatures are both handled by this
-            _inject_args_into_signature(other, **self.injectArgs)
-        return other
-
-    def __ior__(self, other: Union[chain, Signature]) -> Union[chain, Signature]:
-        return self | other
+            _inject_args_into_signature(r, **self.injectArgs)
+        return r
 
 
 def _inject_args_into_signature(sig, **kwargs):
@@ -73,7 +71,10 @@ def _inject_args_into_signature(sig, **kwargs):
 
     existing = task.kwargs.keys()
     subset = {k: v for k, v in kwargs.items() if k not in existing}
-    task.kwargs.update(subset)
+
+    # Don't update kwargs in-place, because that may change the kwargs in the original object
+    # even where the object was cloned in __ior__ above (kwargs isn't copied and is still the same object)
+    task.kwargs = task.kwargs | subset
 
 
 Signature.injectArgs = _inject_args_into_signature
@@ -190,6 +191,11 @@ def set_execution_options(sig: Signature, **options):
         sig.set(**options)
 
 
+def set_use_cache(sig: Signature, use_cache: bool):
+    """Set the :attr:`use_cache` execution option in every task in :attr:`sig`"""
+    set_execution_options(sig, use_cache=use_cache)
+
+
 def set_priority(sig: Signature, priority: int):
     """Set the :attr:`priority` execution option in every task in :attr:`sig`"""
     set_execution_options(sig, priority=priority)
@@ -221,6 +227,7 @@ def get_label(sig: Signature):
 
 Signature.set_execution_options = set_execution_options
 Signature.set_priority = set_priority
+Signature.set_use_cache = set_use_cache
 Signature.set_queue = set_queue
 Signature.set_soft_time_limit = set_soft_time_limit
 Signature.set_label = set_label
