@@ -1,10 +1,10 @@
 import copy
 import time
-from collections import namedtuple
+from collections import namedtuple, deque
 import weakref
 
 from pprint import pformat
-from typing import Union, Iterator
+from typing import Union, Iterator, Optional, Iterable
 
 from celery import current_app
 from celery.result import AsyncResult
@@ -18,6 +18,8 @@ from firexkit.revoke import RevokedRequests
 RETURN_KEYS_KEY = '__task_return_keys'
 _TASK_POST_RUN_KEY = 'TASK_POST_RUN'
 
+ResultId = Union[AsyncResult, str]
+
 logger = get_task_logger(__name__)
 
 
@@ -244,13 +246,13 @@ def _is_worker_alive(result: AsyncResult, retries=1):
                 return True
 
             task_id = result.id
-            task_info = get_task(method_args=(task_id,), destination=(hostname,), timeout=60)
+            task_info = get_task(method_args=(task_id,), destination=(hostname,), timeout=180)
             if task_info and any(task_info.values()):
                 return True
 
             # Try get_active and get_reserved, since we suspect query_task (the api used by get_task above)
             # may be broken sometimes.
-            active_tasks = get_active(destination=(hostname,), timeout=60)
+            active_tasks = get_active(destination=(hostname,), timeout=180)
             task_list = active_tasks.get(hostname) if active_tasks else None
             if task_list:
                 for task in task_list:
@@ -258,7 +260,7 @@ def _is_worker_alive(result: AsyncResult, retries=1):
                     if this_task_id == task_id:
                         return True
 
-            reserved_tasks = get_reserved(destination=(hostname,), timeout=60)
+            reserved_tasks = get_reserved(destination=(hostname,), timeout=180)
             task_list = reserved_tasks.get(hostname) if reserved_tasks else None
             if task_list:
                 for task in task_list:
@@ -283,7 +285,7 @@ def _is_worker_alive(result: AsyncResult, retries=1):
                 logger.debug(f'Queue "{task_queue}" for {task_name} not seen yet; assuming task is alive.')
                 return True
 
-            queues = get_active_queues(timeout=60)
+            queues = get_active_queues(timeout=180)
             active_queues = {queue['name'] for node in queues.values() for queue in node} if queues else set()
             if task_queue in active_queues:
                 return True
@@ -559,7 +561,7 @@ def get_task_results(results: dict) -> dict:
     except KeyError:
         return {}
     else:
-        return {k: results[k] for k in return_keys} if return_keys else {}
+        return {k: results[k] for k in return_keys if k in results} if return_keys else {}
 
 
 def get_tasks_inputs_from_result(results: dict) -> dict:
@@ -573,43 +575,39 @@ def get_tasks_inputs_from_result(results: dict) -> dict:
         return {k: v for k, v in results.items() if k not in return_keys}
 
 
-def _get_results(result: AsyncResult, return_keys_only=True, merge_children_results=False) -> dict:
-    # Generally,
-    #   1.	Apply inputs from current result
-    #   2.	Apply child results (and inputs) from children (if merge_children_results is True)
-    #   3.	Apply output from current result
+def _get_all_results(result: AsyncResult,
+                     all_results: dict,
+                     return_keys_only=True,
+                     merge_children_results=False,
+                     exclude_id=None)-> None:
 
-    results = {}
-    returned_values = {}
     if not result:
-        return results
-    try:
-        if result.successful():
-            _results = copy.deepcopy(result.result) if isinstance(result.result, dict) else result.result
-            if _results:
-                # these will need to be updated only after the children results are extracted
-                # i.e., node results are applied after children results
-                returned_values = get_task_results(_results)
-                if not return_keys_only:
-                    # Only relevant if merge_children_results is True:
-                    #   If return_keys_only us False, we need to apply the inputs first
-                    #   before applying the results(and inputs) of the children
-                    results = get_tasks_inputs_from_result(_results)
+        return  # <-- Nothing to do
 
-        if merge_children_results:
-            children = result.children
-            if children:
-                for child in children:
-                    child_results = get_results(child,
-                                                return_keys_only=return_keys_only,
-                                                merge_children_results=merge_children_results)
-                    results.update(child_results)
+    if result.successful():
+        ret = getattr(result, 'result', {}) or {}
+    else:
+        ret = {}
 
-        results.update(returned_values)
-    except Exception as e:
-        logger.exception(e)
+    if not return_keys_only and ret:
+        # Inputs from child win, below
+        all_results.update(get_tasks_inputs_from_result(ret))
 
-    return results
+    children = getattr(result, 'children', []) or [] if merge_children_results else []
+
+    for child in children:
+        if exclude_id and child and child.id == exclude_id:
+            continue
+        # Beware, recursion
+        _get_all_results(child,
+                         all_results=all_results,
+                         return_keys_only=return_keys_only,
+                         merge_children_results=merge_children_results,
+                         exclude_id=exclude_id) # Unnecessary; exclude_id is usually a first-level child
+
+    if ret:
+        # Returns from the parent win
+        all_results.update(get_task_results(ret))
 
 
 def results2tuple(results: dict, return_keys: Union[str, tuple]) -> tuple:
@@ -652,36 +650,37 @@ def get_results(result: AsyncResult,
         If `return_keys` parameter was specified, returns a tuple of the results in the same order of the return_keys.
         If `return_keys` parameter wasn't specified, return a dictionary of the key/value pairs of the returned results.
     """
-    def _update_results_dict(n):
-        node_results = _get_results(n,
-                                    return_keys_only=return_keys_only,
-                                    merge_children_results=merge_children_results)
-        for k, v in node_results.items():
-            if k not in extracted_dict:
-                # Since we're walking up the chain, children gets precedence in case we get the same key
-                extracted_dict[k] = v
+    all_results = {}
 
-    extracted_dict = {}
+    chain_members = [result]
+    if extract_from_parents:
+        current_node = result
+        while current_node and current_node.id != parent_id and current_node.parent:
+            chain_members.append(current_node.parent)
+            current_node = current_node.parent
 
-    # Get results from current AsyncResult object
-    node = result
-    _update_results_dict(node)
+    while len(chain_members) > 1:
+        # This means we have at least one parent to walk. Parents need to be walked first
+        # because we want the latter services in a chain to override the earlier services
+        # results. But we don't want to walk the child which is a member of the chain,
+        # since this will be walked explicitly, so we exclude that.
+        _get_all_results(result=chain_members.pop(),
+                         all_results=all_results,
+                         return_keys_only=return_keys_only,
+                         merge_children_results=merge_children_results,
+                         exclude_id=chain_members[-1].id)
 
-    # Get results from parents, if applicable
-    if node:
-        stop_at = node.id if extract_from_parents is False else parent_id
-        while node.id != stop_at:
-            node = node.parent
-            if node:
-                _update_results_dict(node)
-            else:
-                break
+    # After possibly walking parents, we get our results for "result" (and possibly all children)
+    _get_all_results(result=result,
+                     all_results=all_results,
+                     return_keys_only=return_keys_only,
+                     merge_children_results=merge_children_results)
 
     from firexkit.task import FireXTask
     if not return_keys or return_keys == FireXTask.DYNAMIC_RETURN or return_keys == (FireXTask.DYNAMIC_RETURN,):
-        return extracted_dict
+        return all_results
     else:
-        return results2tuple(extracted_dict, return_keys)
+        return results2tuple(all_results, return_keys)
 
 
 def get_results_with_default(result: AsyncResult,
@@ -775,3 +774,94 @@ def last_causing_chain_interrupted_exception(ex):
     while e.__cause__ is not None and isinstance(e.__cause__, ChainInterruptedException):
         e = e.__cause__
     return e
+
+
+def climb_up_until_null_parent(result: AsyncResult) -> AsyncResult:
+    r = result
+    while r.parent is not None:
+        r = r.parent
+    return r
+
+
+def get_result_id(r: ResultId) -> str:
+    """Return the string id of r if it was an AsyncResult, otherwise returns r"""
+    return r.id if isinstance(r, AsyncResult) else r
+
+
+def get_all_children(node, timeout=180, skip_subtree_nodes: Optional[list[ResultId]] = None) -> Iterator[AsyncResult]:
+    """Iterate the children of node, skipping any nodes found in skip_subtree_nodes"""
+    stack = deque([node])
+    timeout_time = time.monotonic() + timeout
+    while stack:
+        node = stack.popleft()
+        node_name = get_result_logging_name(node)
+        if skip_subtree_nodes and get_result_id(node) in skip_subtree_nodes:
+            continue
+        yield node
+
+        while time.monotonic() < timeout_time and not node.ready():
+            logger.debug(f'{node_name} state is still {node.state}...sleeping and retrying')
+            time.sleep(0.5)
+        stack.extend(child for child in node.children or [])
+
+
+def forget_single_async_result(r: AsyncResult):
+    """Forget the result of this task
+
+    AsyncResult.forget() also forgets the parent (which is not always desirable), so, we had to implement our own
+    """
+    logger.debug(f'Forgetting result: {get_result_logging_name(r)}')
+    r._cache = None
+    r.backend.forget(r.id)
+
+
+def forget_subtree_results(head_node_result: AsyncResult,
+                           skip_subtree_nodes: Optional[Iterable[ResultId]] = None,
+                           do_not_forget_nodes: Optional[Iterable[ResultId]] = None) -> None:
+    """Forget results of the subtree rooted at head_node_result, while skipping subtrees in skip_subtree_nodes,
+    as well as nodes in do_not_forget_nodes
+    """
+
+    # Must get all the elements from the get_all_children() generator first!
+    # We can't process the forgetting of one element at a time per iteration because the parent/children relationship
+    # might be lost once we forget a node
+    #
+    subtree_nodes = set(get_all_children(head_node_result, skip_subtree_nodes=skip_subtree_nodes))
+
+    do_not_forget_nodes_ids = {get_result_id(n) for n in do_not_forget_nodes} if do_not_forget_nodes else {}
+
+    nodes_to_forget = {n for n in subtree_nodes if n.id not in do_not_forget_nodes_ids}
+
+    msg = [f'Forgetting {len(nodes_to_forget)} results for tree root {get_result_logging_name(head_node_result)}']
+    if skip_subtree_nodes:
+        msg += [f'Skipping subtrees: {[get_result_logging_name(r) for r in skip_subtree_nodes]}']
+    if do_not_forget_nodes_ids:
+        msg += [f'Skipping nodes: {[get_result_logging_name(r) for r in do_not_forget_nodes_ids]}']
+    logger.debug('\n'.join(msg))
+
+    for node in nodes_to_forget:
+        forget_single_async_result(node)
+
+
+def forget_chain_results(result: AsyncResult,
+                         forget_chain_head_node_result: bool = True,
+                         do_not_forget_nodes: Optional[Iterable[AsyncResult]] = None,
+                         **kwargs) -> None:
+    """Forget results of the tree rooted at the "chain-head" of result, while skipping subtrees in skip_subtree_nodes,
+    as well as nodes in do_not_forget_nodes.
+
+    If forget_chain_head_node_result is False (default True), do not forget the head of the result chain
+    """
+    _do_not_forget_nodes = set()
+
+    # Get the head of the result chain, i.e., in chain A|B|C, if result is C, find A
+    head_node_result = climb_up_until_null_parent(result)
+
+    if forget_chain_head_node_result is False:
+        _do_not_forget_nodes.add(head_node_result)
+    if do_not_forget_nodes:
+        _do_not_forget_nodes.update(do_not_forget_nodes)
+
+    forget_subtree_results(head_node_result=head_node_result,
+                           do_not_forget_nodes=_do_not_forget_nodes,
+                           **kwargs)
