@@ -6,8 +6,9 @@ import os
 import textwrap
 from collections import OrderedDict
 import inspect
+from datetime import datetime
 from functools import partial
-from typing import Callable, Iterable, Optional, Union
+from typing import Callable, Iterable, Optional, Union, Any
 from urllib.parse import urljoin
 from copy import deepcopy
 import dataclasses
@@ -24,13 +25,14 @@ from celery.app.task import Task
 from celery.local import PromiseProxy
 from celery.signals import task_prerun, task_postrun, task_revoked
 from celery.utils.log import get_task_logger, get_logger
+
 from firexkit.bag_of_goodies import BagOfGoodies
 from firexkit.argument_conversion import ConverterRegister
 from firexkit.result import get_tasks_names_from_results, wait_for_any_results, \
     RETURN_KEYS_KEY, wait_on_async_results_and_maybe_raise, get_result_logging_name, ChainInterruptedException, \
     ChainRevokedException, last_causing_chain_interrupted_exception, \
     wait_for_running_tasks_from_results, WaitOnChainTimeoutError, get_results, \
-    get_task_name_from_result, first_non_chain_interrupted_exception
+    get_task_name_from_result, first_non_chain_interrupted_exception, forget_chain_results
 from firexkit.resources import get_firex_css_filepath, get_firex_logo_filepath
 from firexkit.firexkit_common import JINJA_ENV
 import time
@@ -38,6 +40,11 @@ import time
 REPLACEMENT_TASK_NAME_POSTFIX = '_orig'
 FIREX_REVOKE_COMPLETE_EVENT_TYPE = 'task-firex-revoked'
 
+REDIS_DB_KEY_FOR_RESULTS_WITH_REPORTS = 'FIREX_RESULTS_WITH_REPORTS'
+REDIS_DB_KEY_PREFIX_FOR_ENQUEUE_ONCE_UID = 'ENQUEUE_CHILD_ONCE_UID_'
+REDIS_DB_KEY_PREFIX_FOR_ENQUEUE_ONCE_COUNT = 'ENQUEUE_CHILD_ONCE_COUNT_'
+REDIS_DB_KEY_PREFIX_FOR_CACHE_ENABLED_UID = 'CACHE_ENABLED'
+
 logger = get_task_logger(__name__)
 
 ################################################################
@@ -45,6 +52,13 @@ logger = get_task_logger(__name__)
 _orig_chain_apply_async__ = _chain.apply_async
 
 
+@dataclasses.dataclass
+class TaskEnqueueSpec:
+    signature: Signature
+    inject_abog: bool = True
+    enqueue_opts: Optional[dict[str, Any]] = None
+
+
 def _chain_apply_async(self: _chain, *args: tuple, **kwargs: dict) -> AsyncResult:
     try:
         chain_depth = 0
@@ -63,6 +77,10 @@ class NotInCache(Exception):
     pass
 
 
+class UidNotInjectedInAbog(Exception):
+    pass
+
+
 class TaskContext:
     pass
 
@@ -127,8 +145,10 @@ def create_collapse_ops(flex_collapse_ops_spec):
 def expand_self_op():
     return {'operation': 'expand', 'targets': ['self']}
 
+
 FLAME_COLLAPSE_KEY = '_default_display'
 
+
 def flame_collapse_formatter(ops, task):
     filled_ops = []
     for op in ops:
@@ -244,6 +264,54 @@ class DictWillNotAllowWrites(dict):
         super().update(*args, **kwargs)
 
 
+def get_cache_enabled_uid_dbkey(cache_key_info: str) -> str:
+    return f'{REDIS_DB_KEY_PREFIX_FOR_CACHE_ENABLED_UID}{cache_key_info}'
+
+
+def get_enqueue_child_once_uid_dbkey(enqueue_once_key: str) -> str:
+    return f'{REDIS_DB_KEY_PREFIX_FOR_ENQUEUE_ONCE_UID}{enqueue_once_key}'
+
+
+def get_enqueue_child_once_count_dbkey(enqueue_once_key: str) -> str:
+    return f'{REDIS_DB_KEY_PREFIX_FOR_ENQUEUE_ONCE_COUNT}{enqueue_once_key}'
+
+
+def get_current_enqueue_child_once_uid_dbkeys(db) -> list[str]:
+    return db.client.keys(get_enqueue_child_once_uid_dbkey('*'))
+
+
+def get_current_cache_enabled_uid_dbkeys(db) -> list[str]:
+    return db.client.keys(get_cache_enabled_uid_dbkey('*'))
+
+
+def get_current_enqueue_child_once_uids(db) -> set[str]:
+    """Returns a set of all task/result ids that were executed with enqueue_once"""
+
+    # First, we need to find all the enqueue_once keys
+    keys = get_current_enqueue_child_once_uid_dbkeys(db)
+    # Then we get the task/result ids stored in those keys
+    return {v.decode() for v in db.mget(keys)}
+
+
+def get_current_cache_enabled_uids(db) -> set[str]:
+    """Returns a set of all task/result ids whose tasks were cache-enabled"""
+
+    # First, we need to find all the cache_enabled keys
+    keys = get_current_cache_enabled_uid_dbkeys(db)
+    # Then we get the task/result ids stored in those keys
+    return {v.decode() for v in db.mget(keys)}
+
+
+def add_task_result_with_report_to_db(db, result_id: str):
+    """Append task id to the list of tasks with reports (e.g. tasks decorated with @email)"""
+    db.client.rpush(REDIS_DB_KEY_FOR_RESULTS_WITH_REPORTS, result_id)
+
+
+def get_current_reports_uids(db) -> set[str]:
+    """Return the list of task/results ids for all tasks with reports (e.g. @email) executed so far"""
+    return {v.decode() for v in db.client.lrange(REDIS_DB_KEY_FOR_RESULTS_WITH_REPORTS, 0, -1)}
+
+
 class FireXTask(Task):
     """
     Task object that facilitates passing of arguments and return values from one task to another, to be used in chains
@@ -451,6 +519,14 @@ class FireXTask(Task):
     def called_as_orig(self):
         return True if self.name.endswith(REPLACEMENT_TASK_NAME_POSTFIX) else False
 
+    def has_report_meta(self) -> bool:
+        """Does this task generate a report (e.g. decorated with @email)?"""
+        return hasattr(self, 'report_meta')
+
+    def add_task_result_with_report_to_db(self):
+        """Maintain a list in the backend of all executed tasks that will generate reports"""
+        return add_task_result_with_report_to_db(self.app.backend,  self.request.id)
+
     @abstractmethod
     def pre_task_run(self, extra_events: Optional[dict] = None):
         """
@@ -466,6 +542,11 @@ class FireXTask(Task):
 
         # Send a custom task-started-info event with the args
         if not self.request.called_directly:
+            if self.has_report_meta():
+                # If the task generates a report, append the task id
+                # to the list in the backend of all executed tasks that  generate reports
+                self.add_task_result_with_report_to_db()
+
             self.send_event('task-started-info',
                             firex_bound_args=convert_to_serializable(bound_args),
                             firex_default_bound_args=convert_to_serializable(default_bound_args),
@@ -548,7 +629,7 @@ class FireXTask(Task):
                     results_list.append('  %d. %s: %r' % (n, k, v))
                     n += 1
             else:
-                results_list.append('  %r' % result)
+                results_list.append(f'  {result!r}')
         if results_list:
             content = 'RETURNS\n' + '\n'.join(results_list)
         logger.debug(banner('COMPLETED: %s' % self.name, ch='*', content=content, length=100),
@@ -633,7 +714,7 @@ class FireXTask(Task):
 
     def _get_cache_key(self):
         # Need a sting hash of name + all_args
-        return str((self.name,) + tuple(sorted(self.all_args.items())))
+        return get_cache_enabled_uid_dbkey(str((self.name,) + tuple(sorted(self.all_args.items()))))
 
     def _cache_get(self, cache_key):
         cached_uuid = self.backend.get(cache_key)
@@ -808,6 +889,15 @@ class FireXTask(Task):
     def abog(self) -> DictWillNotAllowWrites:
         return DictWillNotAllowWrites(_instrumentation_context=self, **{**self.bag, **self.default_bound_args})
 
+    @property
+    def uid(self):
+        try:
+            return self.abog['uid']
+        except KeyError:
+            raise UidNotInjectedInAbog('Please ensure you either inject uid explicitly, '
+                                       'or use either self.enqueue_child, self.enqueue_child_and_get_results, '
+                                       'or self.enqueue_in_parallel')
+
     #######################
     # Enqueuing child tasks
 
@@ -832,6 +922,10 @@ class FireXTask(Task):
         if child_result not in self.context.enqueued_children:
             self.context.enqueued_children[child_result] = {}
 
+    def _remove_enqueued_child(self, child_result):
+        if child_result in self.context.enqueued_children:
+            del(self.context.enqueued_children[child_result])
+
     def _update_child_state(self, child_result, state):
         if child_result not in self.context.enqueued_children:
             self._add_enqueued_child(child_result)
@@ -849,7 +943,63 @@ class FireXTask(Task):
         child_results = self.pending_enqueued_children if pending_only else self.enqueued_children
         self.wait_for_specific_children(child_results=child_results, **kwargs)
 
-    def wait_for_specific_children(self, child_results, **kwargs):
+    def forget_child_result(self,
+                            child_result: AsyncResult,
+                            do_not_forget_report_nodes: bool = True,
+                            do_not_forget_enqueue_once_nodes: bool = True,
+                            do_not_forget_cache_enabled_tasks_results: bool = True,
+                            **kwargs):
+
+        """Forget results of the tree rooted at the "chain-head" of child_result, while skipping subtrees in
+        skip_subtree_nodes, as well as nodes in do_not_forget_nodes.
+
+        If do_not_forget_report_nodes is True (default), do not forget report nodes (e.g. nodes decorated with @email)
+
+        If do_not_forget_enqueue_once_nodes is True (default), do not forget subtrees rooted at nodes that were enqueued
+        with enqueue_once
+
+        If do_not_forget_cache_enabled_tasks_results is True (default), do not forget subtrees rooted at nodes that belong to
+        services with cached=True
+        """
+        logger.debug('Forgetting results')
+
+        skip_subtree_nodes: set[str] = set()
+
+        if do_not_forget_enqueue_once_nodes:
+            enqueue_once_subtree_nodes = get_current_enqueue_child_once_uids(self.backend)
+            if enqueue_once_subtree_nodes:
+                skip_subtree_nodes.update(enqueue_once_subtree_nodes)
+                logger.debug(f'Enqueue once subtree nodes: {enqueue_once_subtree_nodes}')
+
+        if do_not_forget_cache_enabled_tasks_results:
+            cache_enabled_subtree_nodes = get_current_cache_enabled_uids(self.backend)
+            if cache_enabled_subtree_nodes:
+                skip_subtree_nodes.update(cache_enabled_subtree_nodes)
+                logger.debug(f'Cache-enabled  subtree nodes: {cache_enabled_subtree_nodes}')
+
+        report_nodes = None
+        if do_not_forget_report_nodes:
+            report_nodes = get_current_reports_uids(self.backend)
+            if report_nodes:
+                logger.debug(f'Report nodes: {report_nodes}')
+
+        forget_chain_results(child_result,
+                             skip_subtree_nodes=skip_subtree_nodes,
+                             do_not_forget_nodes=report_nodes,
+                             **kwargs)
+        # Since we forget the child, we need to also remove it from the list of enqueued_children
+        self._remove_enqueued_child(child_result)
+
+    def forget_specific_children_results(self, child_results: list[AsyncResult], **kwargs):
+        """Forget results for the explicitly provided child_results"""
+        for child in child_results:
+            self.forget_child_result(child, **kwargs)
+
+    def forget_enqueued_children_results(self, **kwargs):
+        """Forget results for the enqueued children of current task"""
+        self.forget_specific_children_results(self.enqueued_children, **kwargs)
+
+    def wait_for_specific_children(self, child_results, forget: bool = False, **kwargs):
         """Wait for the explicitly provided child_results to run and complete"""
         if isinstance(child_results, AsyncResult):
             child_results = [child_results]
@@ -859,11 +1009,15 @@ class FireXTask(Task):
                 wait_on_async_results_and_maybe_raise(child_results, caller_task=self, **kwargs)
             finally:
                 [self._update_child_state(child_result, self._UNBLOCKED) for child_result in child_results]
+                if forget:
+                    self.forget_specific_children_results(child_results)
 
     def enqueue_child(self, chain: Signature, add_to_enqueued_children: bool = True, block: bool = False,
                       raise_exception_on_failure: bool = None,
                       apply_async_epilogue: Callable[[AsyncResult], None] = None, apply_async_options=None,
-                      **kwargs) -> AsyncResult:
+                      forget: bool = False,
+                      inject_uid: bool = False,
+                      **kwargs) -> Optional[AsyncResult]:
         """Schedule a child task to run"""
 
         if raise_exception_on_failure is not None:
@@ -880,6 +1034,10 @@ class FireXTask(Task):
         if isinstance(chain, InjectArgs):
             return
 
+        # Inject uid whenever possible
+        if inject_uid:
+            chain = InjectArgs(uid=self.uid) | chain
+
         verify_chain_arguments(chain)
         child_result = chain.apply_async(**apply_async_options)
         if apply_async_epilogue:
@@ -894,6 +1052,8 @@ class FireXTask(Task):
             finally:
                 if add_to_enqueued_children:
                     self._update_child_state(child_result, self._UNBLOCKED)
+                if forget:
+                    self.forget_specific_children_results([child_result])
         return child_result
 
     def enqueue_child_and_get_results(self,
@@ -958,6 +1118,7 @@ class FireXTask(Task):
                                    extract_task_returns_only: bool = False,
                                    enqueue_once_key: str = '',
                                    extract_from_parents: bool = True,
+                                   forget: bool = False,
                                    **kwargs) -> Union[tuple, dict]:
         """Apply a ``chain``, and extract results from it.
 
@@ -1001,26 +1162,31 @@ class FireXTask(Task):
                                                      block=True,
                                                      **kwargs)
 
-        return get_results(result_promise,
-                           return_keys=return_keys,
-                           merge_children_results=extract_from_children,
-                           return_keys_only=extract_task_returns_only,
-                           extract_from_parents=extract_from_parents)
+        results = get_results(result_promise,
+                              return_keys=return_keys,
+                              merge_children_results=extract_from_children,
+                              return_keys_only=extract_task_returns_only,
+                              extract_from_parents=extract_from_parents)
+
+        if forget:
+            self.forget_specific_children_results([result_promise])
+
+        return results
 
     def enqueue_child_once(self, *args, enqueue_once_key, block=False, **kwargs):
         """See  :`meth:`enqueue_child_once_and_extract`
         """
-        enqueue_child_once_uid_dbkey = 'ENQUEUE_CHILD_ONCE_UID_' + enqueue_once_key
-        enqueue_child_once_count_dbkey = 'ENQUEUE_CHILD_ONCE_COUNT_' + enqueue_once_key
+        enqueue_child_once_uid_dbkey = get_enqueue_child_once_uid_dbkey(enqueue_once_key)
+        enqueue_child_once_count_dbkey = get_enqueue_child_once_count_dbkey(enqueue_once_key)
 
         if self.request.retries > 0:
             # previous run failed, so we assume (possibly incorrectly!) this child also failed
             # and we decrement the enqueue_child count
-            num_previous_runs = self.backend.decr(enqueue_child_once_count_dbkey)
+            num_previous_runs = self.backend.client.decr(enqueue_child_once_count_dbkey)
             logger.warn(f'enqueue_once called for a retrying task with key {enqueue_once_key}.'
                         f' Number of previous runs: {num_previous_runs+1}. Discounting one previous run.')
 
-        num_runs_attempted = self.backend.incr(enqueue_child_once_count_dbkey)
+        num_runs_attempted = self.backend.client.incr(enqueue_child_once_count_dbkey)
         if int(num_runs_attempted) == 1:
             # This is the first attempt, enqueue the child
             apply_async_epilogue = partial(_set_taskid_in_db_key,
@@ -1100,6 +1266,20 @@ class FireXTask(Task):
             self.wait_for_specific_children(promises, raise_exception_on_failure=raise_exception_on_failure)
         return promises
 
+    def enqueue_child_from_spec(self,
+                                task_spec: TaskEnqueueSpec,
+                                inject_args: Optional[dict] = None):
+        enqueue_opts = task_spec.enqueue_opts or dict()
+        chain = task_spec.signature
+        args_to_inject = self.abog.copy() if task_spec.inject_abog else {}
+        if inject_args:
+            args_to_inject.update(inject_args)
+        if args_to_inject:
+            from firexkit.chain import InjectArgs
+            chain = InjectArgs(**args_to_inject) | chain
+        logger.debug(f'Enqueuing {task_spec}')
+        self.enqueue_child(chain, **enqueue_opts)
+
     def revoke_nonready_children(self):
         nonready_children = self.nonready_enqueued_children
         if nonready_children:
@@ -1446,10 +1626,18 @@ def _custom_serializers(obj) -> str:
 
     return None
 
+
 def convert_to_serializable(obj, max_recursive_depth=10, _depth=0):
+
+    if obj is None or isinstance(obj, (int, float, str, bool)):
+        return obj
+
     if hasattr(obj, 'firex_serializable'):
         return obj.firex_serializable()
 
+    if isinstance(obj, datetime):
+        obj = obj.isoformat()
+
     if dataclasses.is_dataclass(obj):
         try:
             obj = dataclasses.asdict(obj)
@@ -1463,7 +1651,10 @@ def convert_to_serializable(obj, max_recursive_depth=10, _depth=0):
     if _depth < max_recursive_depth:
         # Full object isn't jsonable, but some contents might be. Try walking the structure to get jsonable parts.
         if isinstance(obj, dict):
-            return {k: convert_to_serializable(v, max_recursive_depth, _depth+1) for k, v in obj.items()}
+            return {
+                convert_to_serializable(k, max_recursive_depth, _depth+1): convert_to_serializable(v, max_recursive_depth, _depth+1)
+                for k, v in obj.items()
+            }
 
         # Note that it's important this DOES NOT catch strings, and it won't since strings are jsonable.
         if isinstance(obj, Iterable):
